;;; Copyright (c) 2006 David Lichteblau
;;;
;;; Permission is hereby granted, free of charge, to any person
;;; obtaining a copy of this software and associated documentation files
;;; (the "Software"), to deal in the Software without restriction,
;;; including without limitation the rights to use, copy, modify, merge,
;;; publish, distribute, sublicense, and/or sell copies of the Software,
;;; and to permit persons to whom the Software is furnished to do so,
;;; subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be
;;; included in all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
;;; EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
;;; MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
;;; NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
;;; BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
;;; ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
;;; CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
;;; SOFTWARE.

(in-package :sb-heapdump)

(defvar *dumpload-verbose* t)

;; fixme: can't be deal with fast method plists in dump-package instead
;; of normal fixup routines?
(defvar *fast-methods-kludge*)

(defun load-dumpfile (pathname &key parameters suppress-initializer)
  (when *dumpload-verbose*
    (format t "~&; loading ~A" pathname)
    (force-output))
  (with-open-file (s pathname :element-type :default :external-format :utf8)
    (let* ((length (file-length s))
           (base-sap (sb-sys:int-sap *temporary-space-start*))
	   (*fast-methods-kludge* nil)
	   handle)
      (unless (<= (+ *temporary-space-start* length)
		  *temporary-space-end*)
	(error "heap image length exceeds temporary space size"))
      (sb-posix:mmap base-sap
		     length
		     (logior sb-posix:prot-read sb-posix:prot-write)
		     (logior sb-posix:map-fixed sb-posix:map-private)
		     (sb-sys:fd-stream-fd s)
		     0)
      (sb-sys:without-interrupts
	(sb-sys:without-gcing
	 (let ((header
		(sb-kernel:make-lisp-obj (sb-sys:sap-ref-word base-sap 0))))
	   (when *dumpload-verbose*
	     (write-string " fixup")
	     (force-output))
	   (sb-ext:with-unlocked-packages (:sb-pcl)
	     (apply-fixups base-sap (header-fixups header) parameters))
	   (setf handle (cons header nil))
	   (when *dumpload-verbose*
	     (write-string " gc")
	     (force-output))
	   (let ((sb-kernel:*gc-inhibit* nil))
	     (sb-ext:gc)))) 
	(loop
	    for (fun . plist) in *fast-methods-kludge*
	    do (setf (sb-pcl::method-function-plist fun) plist))
	(sb-posix:munmap base-sap length)
	(when *dumpload-verbose*
	  (format t " done~%"))
	(let ((header (car handle)))
	  (if (and (header-initializer header) (not suppress-initializer))
	      (funcall (header-initializer header) (header-object header))
	      (values (header-object header) (header-initializer header))))))))

(defun apply-fixups (base-sap fixups parameters)
  (dolist (f fixups)
    (let ((value
	   (sb-kernel:get-lisp-obj-address (resolve-fixup f parameters))))
      (loop
	  for location across (fixup-locations f)
	  do (setf (sb-sys:sap-ref-word base-sap location) value)))))

(defun safe-fixup-id (f)
  (when (eql (fixup-id f) (ash #xdeadbeef -2))
    (error "unresolved fixup reference"))
  (fixup-id f))

(defun safe-fixup-id2 (f)
  (when (eql (fixup-id2 f) (ash #xdeadbeef -2))
    (error "unresolved fixup reference"))
  (fixup-id2 f))

(defun resolve-fixup (f parameters)
  (ecase (fixup-type f)
    (#.+parameter-fixup+
      (elt parameters (safe-fixup-id f)))
    (#.+package-fixup+
      (let ((name (safe-fixup-id f)))
	(or (find-package name)
	    (error "referenced package ~S not present" name))))
    (#.+symbol-fixup+
      (intern (safe-fixup-id f) (safe-fixup-id2 f)))
    (#.+classoid-fixup+
      (sb-kernel:find-classoid (safe-fixup-id f)))
    (#.+layout-fixup+
      (sb-kernel:classoid-layout (safe-fixup-id f)))
    (#.+fdefn-fixup+
      (let* ((name (safe-fixup-id f)))
	(or (sb-int:info :function :definition name)
	    (error "referenced function ~S not present" name))))
    (#.+named-type-fixup+
      (let ((result (sb-kernel:values-specifier-type (safe-fixup-id f))))
	(check-type result sb-kernel:named-type)
	result))
    (#.+array-type-fixup+
      (apply #'sb-kernel:make-array-type (safe-fixup-id f)))
    (#.+class-fixup+
      (find-class (safe-fixup-id f)))
    (#.+function-fixup+
      (fdefinition (safe-fixup-id f)))
    (#.+ctor-fixup+
      (destructuring-bind (fn class &rest initargs)
	  (safe-fixup-id f)
	(sb-pcl::ensure-ctor fn class initargs)))
    (#.+slot-accessor-fixup+
      (let ((x (safe-fixup-id f)))
	(sb-pcl::ensure-accessor (fourth x) x (third x))
	(fdefinition x)))
    (#.+fast-method-fixup+
      (push (cons (safe-fixup-id f) (safe-fixup-id2 f)) *fast-methods-kludge*)
      nil)
    (#.+raw-address-fixup+
      (let ((object (safe-fixup-id f)))
	(if (functionp object)
	    (let* ((new-fun
		    (sb-kernel:get-lisp-obj-address
		     (sb-kernel:%closure-fun object))))
	      (setf (object-ref-word object 1)
		    (+ (logandc2 new-fun sb-vm:lowtag-mask)
		       (* sb-vm:simple-fun-code-offset
			  sb-vm:n-word-bytes))))
	    (let* ((new-fun
		    (sb-kernel:get-lisp-obj-address
		     (sb-kernel:fdefn-fun object))))
	      (setf (object-ref-word object 3)
		    (+ (logandc2 new-fun sb-vm:lowtag-mask)
		       (* sb-vm:simple-fun-code-offset
			  sb-vm:n-word-bytes)))))))))
